﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public class VirtualLeader : MonoBehaviour {

    public uint startingFormationIndex = 0;
    public FormationType[] formationTypes;
    public float maxDrift = 1.0f;
    public float predictionTime = 0.5f;

    private NavMeshAgent mAgent;
    private List<SquadMemberBehavior> mMembers;
    private List<Formation> mFormations;
    private float mOriginalSpeed;
    private float mSpeedModifier;
    private int mMaxNumMembers;
    private int mCurrentFormation;

	// Use this for initialization
	void Start ()
	{
        mAgent = GetComponent<NavMeshAgent>();
        mCurrentFormation = (int)startingFormationIndex;
        mOriginalSpeed = mAgent.speed;

        mMembers = new List<SquadMemberBehavior>();
        mFormations = new List<Formation>();
        mSpeedModifier = 1.0f;

        CreateFormations();
	}

	// Update is called once per frame
	void Update () 
	{
        CheckSpeed();
	}

    /// <summary>
    /// Register a unit with this virtual leader to be a part of its formation.
    /// </summary>
    /// <param name="unit"></param>
    /// <returns></returns>
    public bool RegisterMember(SquadMemberBehavior unit)
    {
        if (mMembers.Count == mMaxNumMembers)
        {
            Debug.LogWarning("WARNING: Virtual leader, " + gameObject.name + ", has no open slots!");
            return false;
        }

        if (!mMembers.Exists(u => u.GetInstanceID() == unit.GetInstanceID()))
        {
            mMembers.Add(unit);
            return true;
        }
        else
        {
            Debug.LogWarning("WARNING: Unit, " + unit.name + ", attempted to register with virtual leader twice!");
            return false;
        }
    }
    /// <summary>
    /// Get a specific unit's position in the formation
    /// </summary>
    /// <param name="instanceID"></param>
    /// <returns></returns>
    public Vector3 GetUnitPosition(SquadMemberBehavior unit, out Vector3 targetPosition)
    {
        targetPosition = Vector3.zero;
        Vector3 unitPosition = Vector3.zero;
        Vector3 relativeUnitPosition = Vector3.zero;

        //find index of unit
        int unitIndex = mMembers.FindIndex(u => u.GetInstanceID() == unit.GetInstanceID());

        //get relative position
        if (unitIndex >= 0)
            relativeUnitPosition = mFormations[mCurrentFormation].GetUnitPosition(unitIndex);
        else
        {
            Debug.LogWarning("WARNING: " + unit.name + " is requesting position from " + gameObject.name + " and is not a member!");
            return unitPosition;
        }

        //get formation position
        unitPosition = transform.TransformPoint(relativeUnitPosition);

        //get prediction position
        float dist = mAgent.speed * predictionTime;
        NavMeshHit prediction;
        mAgent.SamplePathPosition(1, dist, out prediction);
        targetPosition = transform.TransformPoint(transform.InverseTransformPoint(prediction.position) + relativeUnitPosition);

        return unitPosition;
    }
    /// <summary>
    /// Tells virtual leader to begin moving toward a target position
    /// </summary>
    /// <param name="target"></param>
    public void MoveTo(Vector3 target)
    {
        mAgent.SetDestination(target);
    }

    /// <summary>
    /// Create necessary formations and determine the maximum number of members
    /// </summary>
    private void CreateFormations()
    {
        foreach(FormationType type in formationTypes)
        {
            if (!mFormations.Exists(f => f.Type == type))
            {
                switch (type)
                {
                    case FormationType.BASIC_WEDGE:
                        mFormations.Add(new BasicWedgeFormation());
                        break;
                    default:
                        Debug.LogError("ERROR: Formation Type requested for " + gameObject.name + " not implemented!");
                        break;
                }
            }
            else
                Debug.LogWarning("WARNING: Duplicate formation requested for " + gameObject.name + "!");
        }

        mMaxNumMembers = mFormations[0].MaxUnits;

        foreach(Formation formation in mFormations)
        {
            if (formation.MaxUnits < mMaxNumMembers)
                mMaxNumMembers = formation.MaxUnits;
        }
    }
    /// <summary>
    /// Modifies the leader's speed to keep it from getting ahead of the formation
    /// </summary>
    private void CheckSpeed()
    {
        if (mMembers.Count > 0)
        {
            //calculate squad's center of mass
            Vector3 centerOfMass = Vector3.zero;
            foreach (SquadMemberBehavior member in mMembers)
            {
                centerOfMass += member.transform.position;
            }
            centerOfMass /= mMembers.Count;

            float distFromCM = (centerOfMass - transform.position).sqrMagnitude - mFormations[mCurrentFormation].CenterOfMassOffsetSqr; //distance from center of mass

            mSpeedModifier = (maxDrift * maxDrift - distFromCM) / (maxDrift * maxDrift);

            mAgent.speed = mOriginalSpeed * mSpeedModifier; //calculate new speed based on modifier
        }
    }
}
